package engine;

import flash.display.BitmapData;
import flash.geom.Rectangle;
import flash.geom.Point;

/**
 * ...
 * @author randomnine
 */

class FontSheet 
{
	var m_fontBD:BitmapData;

	var m_charWidth:Int;
	var m_charHeight:Int;
	var m_charSpacing:Int;
	var m_spaceWidth:Int;
	
	var m_charWidths:Array<Int>;

	public function new( _fontBD:BitmapData, _charWidth:Int, _charHeight:Int, _charSpacing:Int, _spaceWidth:Int ) 
	{
		m_fontBD = _fontBD;
		m_charWidth = _charWidth;
		m_charHeight = _charHeight;
		m_charSpacing = _charSpacing;
		m_spaceWidth = _spaceWidth;
		
		m_charWidths = new Array<Int>();
		
		for ( i in 0 ... 128 )
		{
			var charX = getCharX(i);
			var charY = getCharY(i);
			
			var testWidth = m_charWidth;
			while ( testWidth > 1 )
			{
				var foundPixel:Bool = false;
				
				for ( y in 0...m_charHeight )
				{
					if ( (m_fontBD.getPixel32( charX + testWidth - 1, charY + y ) & 0xFF000000) > 0x08000000
					  || (m_fontBD.getPixel32( charX + testWidth - 1, charY + y ) & 0xFF000000) < 0 ) // Signed math. ¬_¬
					{
						foundPixel = true;
						break;
					}
				}
				
				if ( foundPixel )
				{
					break;
				}
				
				testWidth--;
			}
			
			m_charWidths[i] = testWidth;
		}
		
		m_charWidths[32] = m_spaceWidth - m_charSpacing;
	}
	
	public function getCharX( _charCode:Int )
	{
		return m_charWidth * (_charCode % 16);
	}
	
	public function getCharY( _charCode:Int )
	{
		return m_charHeight * Std.int( _charCode / 16 );
	}
	
	public function renderText( _targetBD:BitmapData, _text:String, _x:Int, _y:Int, _xBias:Float=1.0 )
	{
		var startX:Int = _x;
		
		if ( _xBias < 1.0 )
		{
			// Calculate text width.
			for ( i in 0 ... _text.length )
			{
				var charCode:Int = _text.charCodeAt(i);
				var charWidth:Int = m_charWidths[charCode];

				 startX -= charWidth + m_charSpacing;
			}
			
			 startX = Std.int(  startX + (_x -  startX) * _xBias );
		}

		var curX:Int = startX;
		var curY:Int = _y;
		
		for ( i in 0 ... _text.length )
		{
			var charCode:Int = _text.charCodeAt(i);
			var charWidth:Int = m_charWidths[charCode];
			
			var glyphRect:Rectangle = new Rectangle( getCharX(charCode), getCharY(charCode), charWidth, m_charHeight );
			
			_targetBD.copyPixels( m_fontBD, glyphRect, new Point( curX, curY ) );
			
			if ( charCode == 13 || charCode == 10 )
			{
				curX = startX;
				curY += 9;
			}
			else
			{
				curX += charWidth + m_charSpacing;
			}
		}
	}
}
