package engine;

import flash.events.KeyboardEvent;
import flash.Lib;

/**
 * ...
 * @author randomnine
 */

class KeyState
{
	public var m_down:Bool;
	public var m_changed:Bool;

	public function new()
	{
		m_down = false;
		m_changed = false;
	}
}

class GameInput 
{
	public static inline var c_keyArrowLeft : UInt = 37;
	public static inline var c_keyArrowUp : UInt = 38;
	public static inline var c_keyArrowRight : UInt = 39;
	public static inline var c_keyArrowDown : UInt = 40;
	
	public static inline var c_keyA : UInt = 65;
	public static inline var c_keyD : UInt = 68;
	public static inline var c_keyM : UInt = 77;
	public static inline var c_keyQ : UInt = 81;
	public static inline var c_keyR : UInt = 82;
	public static inline var c_keyS : UInt = 83;
	public static inline var c_keyV : UInt = 86;
	public static inline var c_keyW : UInt = 87;
	public static inline var c_keyX : UInt = 88;
	public static inline var c_keyZ : UInt = 90;
	
	var m_keyStates:Array<KeyState>;

	private var m_mouseX : Float;
	private var m_mouseY : Float;
	private var m_mouseClickPending : Bool;
	private var m_mouseClickActive : Bool;

	var m_pendingEvents:List<KeyboardEvent>;
	
	public function new() 
	{
		// Initialise keystates.
		m_keyStates = new Array<KeyState>();
		for ( i in 0...256 )
		{
			m_keyStates[i] = new KeyState();
		}
		
		m_pendingEvents = new List<KeyboardEvent>();
		
		Lib.current.stage.addEventListener( KeyboardEvent.KEY_DOWN, onKeyDown );
		Lib.current.stage.addEventListener( KeyboardEvent.KEY_UP, onKeyUp );
	}

	public function onKeyDown( _event:KeyboardEvent )
	{
		m_pendingEvents.add( _event );
	}

	public function onKeyUp( _event:KeyboardEvent )
	{
		m_pendingEvents.add( _event );
	}
	
	public function notifyMouseMove( _x:Float, _y:Float, _clicked:Bool )
	{
		m_mouseX = _x;
		m_mouseY = _y;
		m_mouseClickPending = _clicked;
	}
	
	public function preFrameAdvance()
	{
		for ( elem in m_keyStates )
		{
			elem.m_changed = false;
		}
		
		var event:KeyboardEvent;
		while ( (event = m_pendingEvents.pop()) != null )
		{
			var keyState:KeyState = m_keyStates[event.keyCode];
			
			var shouldBeDown:Bool = ( event.type == KeyboardEvent.KEY_DOWN );
			
			if ( keyState.m_down != shouldBeDown )
			{
				keyState.m_down = shouldBeDown;
				keyState.m_changed = true;
			}
		}
		
		m_mouseClickActive = m_mouseClickPending;
		m_mouseClickPending = false;
	}

	public function isKeyDown( _code:Int ) : Bool
	{
		return m_keyStates[_code].m_down;
	}

	public function wasKeyPressed( _code:Int ) : Bool
	{
		return m_keyStates[_code].m_down && m_keyStates[_code].m_changed;
	}

	public function wasKeyReleased( _code:Int ) : Bool
	{
		return !m_keyStates[_code].m_down && m_keyStates[_code].m_changed;
	}
	
	public function getMouseX() : Float { return m_mouseX; }
	public function getMouseY() : Float { return m_mouseY; }
	public function getMouseClicked() : Bool { return m_mouseClickActive; }
}
