package engine;

import flash.display.Bitmap;
import flash.display.BitmapData;
import flash.display.LoaderInfo;
import flash.events.Event;
import flash.events.MouseEvent;
import flash.events.TimerEvent;
import flash.geom.Point;
import flash.Lib;
import flash.utils.Timer;
import game.data.FontData;

import game.GameSpec;

/**
 * ...
 * @author randomnine
 */

class Main 
{
	private static inline var c_frameTimeMS:Int = 25;
	
	var m_canvasBD:BitmapData;
	var m_canvasBitmap:Bitmap;

	var m_overlayBD:BitmapData;
	var m_overlayBitmap:Bitmap;
	
	var m_pauseBD:BitmapData;
	var m_pauseBitmap:Bitmap;

	var m_sceneManager:GameSceneManager;
	var m_pauseScene:GameScene;
	
	var m_gameTimer:Timer;
	var m_lastFrameTime:Int;
	
	var m_advanceContext:AdvanceContext;
	var m_gameInput:GameInput;

	var m_allowPause:Bool;
	var m_gameHasFocus:Bool;
	
	var m_downloadProgress:Float;
	var m_downloadComplete:Bool;
	
	var m_stutter:Bool;
	
    static function main()
    {
		var main:Main = new Main();
    }
	
	function new()
	{
		// Set up the canvas.
		m_canvasBD = new BitmapData( GameSpec.c_width, GameSpec.c_height, false, 0x000000 );
		m_canvasBitmap = new Bitmap( m_canvasBD );
		m_canvasBitmap.scaleX = m_canvasBitmap.scaleY = GameSpec.c_pixelScale;
		
		Lib.current.addChild( m_canvasBitmap );
		
		m_overlayBD = null;
		m_overlayBitmap = null;
		
		m_pauseBD = new BitmapData( GameSpec.c_width, GameSpec.c_height, true, 0x00000000 );
		m_pauseBitmap = new Bitmap( m_pauseBD );
		m_pauseBitmap.scaleX = m_pauseBitmap.scaleY = GameSpec.c_pixelScale;
		
		m_pauseBitmap.visible = false;
		Lib.current.addChild( m_pauseBitmap );
		
		// Kick off our scene manager.
		m_sceneManager = new GameSceneManager( GameSpec.getStartingScene() );
		m_pauseScene = GameSpec.getPauseOverlayScene();

		// Ready all the little things.
		m_advanceContext = new AdvanceContext();
		
		m_advanceContext.m_main = this;
		
		m_gameInput = new GameInput();
		m_advanceContext.m_input = m_gameInput;
		
		// Set up the advance loop and go!
		m_gameTimer = new Timer(1);
		m_gameTimer.addEventListener( TimerEvent.TIMER, onTimerEvent );
		m_gameTimer.start();

		// (Preloader stuff.)
		m_downloadProgress = 0;
		m_downloadComplete = false;
	
		m_lastFrameTime = Lib.getTimer();
		
		// Check for gaining and losing focus.
		Lib.current.stage.addEventListener(Event.ACTIVATE,onActivate);
		Lib.current.stage.addEventListener(Event.DEACTIVATE, onDeactivate);
		m_allowPause = false;
		m_gameHasFocus = true;
		
		// Set up the mouse listener.
		Lib.current.stage.addEventListener( MouseEvent.MOUSE_MOVE, mouseMove );
		Lib.current.stage.addEventListener( MouseEvent.MOUSE_DOWN, mouseClick );
		Lib.current.stage.addEventListener( MouseEvent.MOUSE_UP, mouseMove );
	}
	
	function advanceLoop( _render:Bool )
	{
		m_gameInput.preFrameAdvance();

		// Toggle mute.
		if ( m_gameInput.wasKeyPressed( GameInput.c_keyM ) )
		{
			SoundPlayer.get().setSoundEnabled( !SoundPlayer.get().getSoundEnabled() );
		}
		
		if ( m_gameHasFocus || !m_allowPause )
		{
			m_stutter = !_render;
			m_sceneManager.advance( 0.001 * c_frameTimeMS, m_advanceContext );
			m_pauseBitmap.visible = false;

			if ( _render ) {
				m_canvasBD.lock();
				m_sceneManager.render( m_canvasBD );
				m_canvasBD.unlock();
			}
		}
		else
		{
			m_stutter = !_render;
			m_pauseScene.advance( 0.001 * c_frameTimeMS, m_advanceContext );
			m_pauseBitmap.visible = true;
			Lib.current.setChildIndex( m_pauseBitmap, Lib.current.numChildren - 1 );
			
			if ( _render ) {
				m_canvasBD.lock();
				m_sceneManager.render( m_canvasBD );
				m_pauseScene.render( m_pauseBD );
				m_canvasBD.unlock();
			}
		}
	}

	function onTimerEvent( _e:TimerEvent )
	{
		if ( Lib.getTimer() >= m_lastFrameTime + c_frameTimeMS )
		{
			// Advance preload.
			if ( !m_downloadComplete )
			{
				if ( Lib.current.root.loaderInfo.bytesLoaded >= Lib.current.root.loaderInfo.bytesTotal )
				{
					m_downloadComplete = true;
					m_downloadProgress = 1.0;
					
					Lib.current.gotoAndStop( 2 );
				}
				else
				{
					m_downloadProgress = Lib.current.root.loaderInfo.bytesLoaded / Lib.current.root.loaderInfo.bytesTotal;
				}
			}
			
			while ( Lib.getTimer() >= m_lastFrameTime + 2 * c_frameTimeMS ) {
				// Skip a render frame (run two advance steps).
				m_lastFrameTime += c_frameTimeMS;
				advanceLoop(false);
			}
			
			m_lastFrameTime += c_frameTimeMS;
			advanceLoop(true);
			
			// If we fall behind real time, jump forwards to just running 4 frames quickly.
			m_lastFrameTime = Std.int( Math.max( m_lastFrameTime, Lib.getTimer()-4*c_frameTimeMS ) );
			
			// Redraw the scene after each refresh like this.
			_e.updateAfterEvent();
		}
	}
	
	function onActivate( event:Event )
	{
		m_gameHasFocus = true;
		SoundPlayer.get().onUnpause();
	}

	function onDeactivate( event:Event )
	{
		m_gameHasFocus = false;
		SoundPlayer.get().onPause();
	}
	
	function mouseMove( event:MouseEvent )
	{
		m_gameInput.notifyMouseMove( event.stageX / game.GameSpec.c_pixelScale, event.stageY / game.GameSpec.c_pixelScale, false );
	}

	function mouseClick( event:MouseEvent )
	{
		m_gameInput.notifyMouseMove( event.stageX / game.GameSpec.c_pixelScale, event.stageY / game.GameSpec.c_pixelScale, true );
	}	
	
	public function hasFocus() : Bool { return m_gameHasFocus; }
	public function isDownloadComplete() : Bool { return m_downloadComplete; }
	public function getDownloadProgress() : Float { return m_downloadProgress; }
	
	public function setAllowPause( _allow:Bool )
	{
		m_allowPause = _allow;
	}
	
	public function setOverlay( _overlayBD:BitmapData )
	{
		if ( m_overlayBD == null && _overlayBD != null ) // Create overlay.
		{
			m_overlayBD = _overlayBD;
			m_overlayBitmap = new Bitmap( m_overlayBD );
			Lib.current.addChild( m_overlayBitmap );
		}
		else if ( m_overlayBD != null && _overlayBD == null ) // Destroy overlay.
		{
			Lib.current.removeChild( m_overlayBitmap );
			m_overlayBitmap = null;
		}
		
		// Set overlay.
		m_overlayBD = _overlayBD;
		
		if ( m_overlayBitmap != null )
		{
			m_overlayBitmap.bitmapData = m_overlayBD;
		}
	}
	
	public function getOverlay() : BitmapData
	{
		return m_overlayBD;
	}	
	
	public function isStuttering() : Bool
	{
		return m_stutter;
	}
}
