package engine;
import flash.events.Event;
import flash.media.Sound;
import flash.media.SoundChannel;
import flash.media.SoundTransform;

/**
 * ...
 * @author randomnine
 */

class SoundPlayer 
{
	// Singleton. -------------------------------------------------------
	static private var s_singleton : SoundPlayer;
	
	static public function get() : SoundPlayer
	{
		if ( s_singleton == null )
		{
			s_singleton = new SoundPlayer();
		}
		
		return s_singleton;
	}

	// Object. ----------------------------------------------------------
	var m_soundEnabled : Bool;
	var m_musicEnabled : Bool;
	var m_paused : Bool;
	var m_gameHasFocus : Bool;
	
	var m_musicTrack : Sound;
	var m_musicLoops : Bool;
	var m_musicFreezeOnPause : Bool;
	
	var m_musicChannel : SoundChannel;
	var m_musicVolume : Float;
	var m_musicResumePoint : Float;
	
	public function new() 
	{
		m_soundEnabled = true;
		m_musicEnabled = true;
		m_paused = false;
		m_gameHasFocus = true;
	
		m_musicTrack = null;
		m_musicChannel = null;
		m_musicVolume = 1.0;
	}
	
	public function setSoundEnabled( _enabled:Bool )
	{
		m_soundEnabled = _enabled;
		updateMusicVolume();
	}

	public function setMusicEnabled( _enabled:Bool )
	{
		m_musicEnabled = _enabled;
		updateMusicVolume();
	}

	public function setPaused( _paused:Bool )
	{
		m_paused = _paused;
		updateMusicVolume();
	}

	public function playSound( _sound:Sound, _volume:Float )
	{
		if ( !isPaused() && m_soundEnabled )
		{
			var soundTransform:SoundTransform = new SoundTransform( _volume );
		
			_sound.play( 0, 1, soundTransform );
		}
	}
	
	public function playMusic( _track:Sound, _start:Float, _volume:Float, _isLooping:Bool, _freezeOnPause:Bool )
	{
		stopMusic();

		m_musicVolume = _volume;

		m_musicTrack = _track;
		m_musicLoops = _isLooping;
		m_musicFreezeOnPause = _freezeOnPause;
		
		m_musicChannel = m_musicTrack.play( _start, m_musicLoops ? 10000 : 0, new SoundTransform( (!m_paused && m_soundEnabled && m_musicEnabled) ? m_musicVolume : 0.0 ) );
		m_musicChannel.addEventListener( Event.SOUND_COMPLETE, onMusicEnd );
		
		m_musicFreezeOnPause = _freezeOnPause;
	}
	
	public function onPause()
	{
		m_gameHasFocus = false;
		
		if ( m_musicChannel != null )
		{
			if ( m_musicFreezeOnPause )
			{
				m_musicResumePoint = m_musicChannel.position;
				m_musicChannel.removeEventListener( Event.SOUND_COMPLETE, onMusicEnd );
				m_musicChannel.stop();
				m_musicChannel = null;
			}
			else
			{
				updateMusicVolume();
			}
		}
	}
	
	public function onUnpause()
	{
		m_gameHasFocus = true;
		
		if ( m_musicChannel == null && m_musicTrack != null && m_musicFreezeOnPause )
		{
			m_musicChannel = m_musicTrack.play( m_musicResumePoint + 0.1, m_musicLoops ? 10000 : 0, new SoundTransform( (!m_paused && m_soundEnabled && m_musicEnabled) ? m_musicVolume : 0.0 ) );
			m_musicChannel.addEventListener( Event.SOUND_COMPLETE, onMusicEnd );
		}
		else
		{
			updateMusicVolume();
		}
	}
	
	public function stopMusic()
	{
		if ( m_musicChannel != null )
		{
			m_musicChannel.removeEventListener( Event.SOUND_COMPLETE, onMusicEnd );
			m_musicChannel.stop();
			m_musicChannel = null;

			m_musicTrack = null;
		}
	}
	
	public function setMusicVolume( _volume:Float )
	{
		m_musicVolume = _volume;
		
		updateMusicVolume();
	}
	
	function onMusicEnd( e:Event )
	{
		m_musicChannel = null;
		m_musicTrack = null;
	}
	
	function updateMusicVolume()
	{
		if ( m_musicChannel != null )
		{
			m_musicChannel.soundTransform = new SoundTransform( (!isPaused() && m_soundEnabled && m_musicEnabled) ? m_musicVolume : 0.0 );
		}
	}
	
	public function isPaused() : Bool { return m_paused || !m_gameHasFocus; }
	
	public function getSoundEnabled() : Bool { return m_soundEnabled; }
	public function getMusicEnabled() : Bool { return m_musicEnabled; }
	
	public function isMusicPlaying() : Bool { return m_musicChannel != null; }
	public function getMusicPosition() : Float { return (m_musicChannel != null) ? m_musicChannel.position : 0; }
	public function getMusicVolume() : Float { return m_musicVolume; }
}
