package engine;

import flash.display.BitmapData;
import flash.geom.ColorTransform;

/**
 * ...
 * @author randomnine
 */

class SpriteAnimating
{
	var m_spriteSheet : SpriteSheet;
	var m_variation : Int;
	
	var m_currentAnimation : Int;
	var m_currentFrame : Int;
	var m_frameTime : Float;
	
	var m_dx : Int;
	var m_dy : Int;
	
	var m_hitFrame : Bool;
	var m_changedFrame : Bool;
	var m_animationEnded : Bool;
	
	public function new( _spriteSheet:SpriteSheet ) 
	{
		m_spriteSheet = _spriteSheet;
		m_variation = -1;
		
		m_currentFrame = 0;
		m_frameTime = 0.0;
		
		m_dx = 0;
		m_dy = 0;

		m_changedFrame = false;
		m_animationEnded = false;
	}
	
	public function setVariation( _variation:Int )
	{
		m_variation = _variation;
	}
	
	public function startAnimation( _animation:Int, _delay:Float = 0.0 )
	{
		m_currentAnimation = _animation;
		m_currentFrame = 0;
		m_frameTime = -_delay;
		
		m_changedFrame = true;
		m_animationEnded = false;
		
		m_dx = m_spriteSheet.getAnimation( _animation ).getFrame( 0 ).m_dx;
		m_dy = m_spriteSheet.getAnimation( _animation ).getFrame( 0 ).m_dy;
	}
	
	public function advance( _dt:Float )
	{
		var anim:SpriteSheetAnimation = m_spriteSheet.getAnimation( m_currentAnimation );
		var maxFrameTime = anim.getFrame( m_currentFrame ).m_length;
		
		m_frameTime += _dt;
		
		m_changedFrame = false;
		
		if ( m_frameTime > maxFrameTime )
		{
			if ( m_currentFrame == anim.getLength() - 1 )
			{
				if ( anim.isLooping() )
				{
					m_changedFrame = true;
					
					m_frameTime -= maxFrameTime;
					m_currentFrame = 0;
				}
				else
				{
					m_animationEnded = true;
				}
			}
			else
			{
				m_changedFrame = true;
				
				m_frameTime -= maxFrameTime;
				m_currentFrame++;
			}
		}
		
		if ( m_changedFrame )
		{
			var frame:SpriteSheetAnimationFrame = anim.getFrame( m_currentFrame );
			
			m_dx = frame.m_dx;
			m_dy = frame.m_dy;
		}
	}
	
	public function drawCurrentFrame( _targetBD:BitmapData, _x:Int, _y:Int, _flip:Bool, _transform:ColorTransform )
	{
		var curAnim:SpriteSheetAnimation = m_spriteSheet.getAnimation( m_currentAnimation );
		var curFrame:Int = curAnim.getFrame( m_currentFrame ).m_frame;
		
		m_spriteSheet.drawFrame( _targetBD, m_variation, curFrame, _x, _y, _flip, _transform );
	}
	
	public function getMotionX() : Int { return m_dx; }
	public function getMotionY() : Int { return m_dy; }
	
	public function inHitFrame() : Bool
	{
		return m_spriteSheet.getAnimation( m_currentAnimation ).getFrame( m_currentFrame ).m_hitFrame;
	}
	
	public function hasChangedFrame() : Bool
	{
		return m_changedFrame;
	}
	
	public function hasAnimationEnded() : Bool
	{
		return m_animationEnded;
	}
}
