package engine;
import flash.display.BitmapData;
import flash.geom.ColorTransform;
import flash.geom.Point;
import flash.geom.Rectangle;

/**
 * ...
 * @author randomnine
 */

class SpriteSheet
{
	var m_sheetBD : BitmapData;
	var m_flipBD : BitmapData;
	
	var m_variationBDs : Array<BitmapData>;
	
	var m_frameWidth : Int;
	var m_frameHeight : Int;
	
	var m_framesAcross : Int;
	var m_framesDown : Int;
	
	var m_animations : Array<SpriteSheetAnimation>;
	
	public function new( _sheetBD:BitmapData, _frameWidth:Int, _frameHeight:Int )
	{
		m_sheetBD = _sheetBD;
		m_frameWidth = _frameWidth;
		m_frameHeight = _frameHeight;
		
		m_framesAcross = Std.int( m_sheetBD.rect.width / m_frameWidth );
		m_framesDown = Std.int( m_sheetBD.rect.height / m_frameHeight );
		
		m_animations = new Array<SpriteSheetAnimation>();
		
		// Create the flipped bitmap.
		m_flipBD = new BitmapData( m_sheetBD.width, m_sheetBD.height, true );
		
		var flipRect:Rectangle = new Rectangle( 0, 0, 1, m_sheetBD.height );
		var destPoint:Point = new Point( 0, 0 );
		
		for ( x in 0 ... m_sheetBD.width )
		{
			var frameX:Int = m_frameWidth * Std.int(x / m_frameWidth);
			var subFrameX:Int = x - frameX;
			flipRect.x = frameX + (m_frameWidth-1 - subFrameX);
			destPoint.x = x;
			
			m_flipBD.copyPixels( m_sheetBD, flipRect, destPoint );
		}
		
		m_variationBDs = new Array<BitmapData>();
	}
	
	public function addVariation( _index:Int, _r1:Int, _r2:Int, _g1:Int, _g2:Int, _b1:Int, _b2:Int )
	{
		m_variationBDs[2*_index] = m_sheetBD.clone();
		m_variationBDs[2*_index + 1] = m_flipBD.clone();
		
		for ( x in 0 ... m_sheetBD.width )
		{
			for ( y in 0 ... m_sheetBD.height )
			{
				for ( sheet in 2 * _index ... 2 * _index + 2 )
				{
					var pixel:UInt = m_variationBDs[sheet].getPixel32( x, y );
					
					if ( pixel & 0xff000000 == 0xff000000 ) // Solid alpha.
					{
						var pixel24:UInt = pixel & 0x00ffffff;
						if ( pixel24 == 0xff0000 )
						{
							m_variationBDs[sheet].setPixel( x, y, _r1 );
						}
						else if ( pixel24 == 0x800000 )
						{
							m_variationBDs[sheet].setPixel( x, y, _r2 );
						}
						else if ( pixel24 == 0x00ff00 )
						{
							m_variationBDs[sheet].setPixel( x, y, _g1 );
						}
						else if ( pixel24 == 0x008000 )
						{
							m_variationBDs[sheet].setPixel( x, y, _g2 );
						}
						else if ( pixel24 == 0x0000ff )
						{
							m_variationBDs[sheet].setPixel( x, y, _b1 );
						}
						else if ( pixel24 == 0x000080 )
						{
							m_variationBDs[sheet].setPixel( x, y, _b2 );
						}
					}
				}
			}
		}
	}
	
	public function getNumVariations() : Int { return Std.int( m_variationBDs.length / 2 ); }
	
	public function addAnimation( _index:Int, _animation:SpriteSheetAnimation )
	{
		m_animations[_index] = _animation;
	}
	
	public function drawFrame( _targetBD:BitmapData, _variation:Int, _frame:Int, _x:Int, _y:Int, _flip:Bool, _transform:ColorTransform )
	{
		var sheet:BitmapData = (_variation == -1) ? (_flip ? m_flipBD : m_sheetBD) : m_variationBDs[2 * _variation + (_flip ? 1 : 0)];
		
		var frameX:Int = _frame % m_framesAcross;
		var frameY:Int = Std.int( _frame / m_framesAcross );
		
		if ( frameY < m_framesDown )
		{
			var frameRect:Rectangle = new Rectangle( frameX * m_frameWidth, frameY * m_frameHeight, m_frameWidth, m_frameHeight );
			if ( _transform != null )
			{
				var alphaBD = new BitmapData( m_frameWidth, m_frameHeight, true, 0x00000000 );
				alphaBD.copyPixels( sheet, frameRect, new Point( 0, 0 ) );
				
				alphaBD.colorTransform( alphaBD.rect, _transform );
				
				_targetBD.copyPixels( alphaBD, alphaBD.rect, new Point( _x, _y ) );
			}
			else
			{
				_targetBD.copyPixels( sheet, frameRect, new Point( _x, _y ) );
			}
		}
	}
	
	public function getAnimation( _index:Int ) : SpriteSheetAnimation
	{
		return m_animations[_index];
	}
}
