package game.scenes;

import flash.display.BitmapData;
import flash.display.Shader;
import flash.filters.ShaderFilter;
import flash.geom.Point;
import flash.geom.Rectangle;
import flash.media.Sound;
import flash.utils.ByteArray;

import engine.AdvanceContext;
import engine.GameInput;
import engine.GameScene;
import engine.SoundPlayer;

import game.data.FontData;

class WavApple extends Sound {public function new(){super();}}
class WavCannibal extends Sound {public function new(){super();}}
class WavWall extends Sound {public function new(){super();}}

class FilterDarkblur extends ByteArray {public function new(){super();}}

class SceneSnake implements GameScene
{
	public static inline var c_tileSize = 10;
	public static inline var c_width = 18;
	public static inline var c_height = 13;
	
	public static var s_highScore = 0;
	
	public static var s_blur:Bool = true;
	
	// The blur filter.
	var m_filter:ShaderFilter;
	
	// m_screen is an inline 18x13 array containing the contents of each grid cell.
	// -5 = wall
	// -4 = blood trail
	// -3 = a bloody mess
	// -2 = a dead body part
	// -1 = an apple
	//  0 = empty
	// >0 = live snake parts, numbered by age. 1 is the head, m_snakeLength-1 is the tail.
	var m_screen:Array<Int>;
	
	// "Free spaces" is the number of 0 or -4 cells in m_screen. It's kept up to date so
	// we can efficiently spawn apples in random places.
	var m_freeSpaces:Int;
	
	// X and Y co-ordinates of the snake's head.
	var m_headX:Int;
	var m_headY:Int;
	
	// Length of the snake.
	var m_snakeLength:Int;
	
	// How many frames should we continue bleeding horribly from our mutilated tail for?
	var m_blood:Int;
	
	var m_prevDir:Int; // Stored direction moved last advance frame.
	var m_dir:Int; // Current direction. 0 = right, 1 = down, 2 = left, 3 = up
	var m_setDir:Bool; // Have we set the direction for the next advance frame yet?
	var m_queuedDir:Int; // Allow up to one direction input queued in advance.
	
	var m_nonAdvanceFrames:Int; // Count off non-advance frames between every advance frame.
	
	var m_score:Int;

	var m_gameStart:Bool;
	var m_gameOver:Bool;
	var m_gameOverTime:Float;
	
	public function new() 
	{
		m_filter = new ShaderFilter( new Shader( new FilterDarkblur() ) );
		
		m_screen = new Array<Int>();
		
		for ( i in 0 ... c_width * c_height ) {
			m_screen.push( 0 );
		}

		m_freeSpaces = c_width * c_height;
		
		// Fill in a bunch of points around the edge. This means you can't just loop
		// around the edge of the map easily.
		
		// We want to spawn 8 wall blocks, but they have to be at least 3 blocks apart
		// from each other so as not to create areas you can't access. They can't be
		// a space out from the corner, either.
		for ( i in 0 ... 8 ) {
			while ( true ) {
				// Pick a random spot. (Not very efficient, but this is a one-off and
				// doesn't happen during gameplay - so who cares!)
				var idx:Int = Math.floor( Math.random() * c_width * c_height );
				var x:Int = idx % c_width;
				var y:Int = Math.floor( idx / c_width );
				
				// Is it clear?
				if ( m_screen[idx] == 0 ) {
					// Yes! Is it along an edge, but not in a corner?
					if ( ((y < 2 || y > c_height - 2) && x >= 2 && x < c_width - 2)
					  || ((x < 2 || x > c_width - 2) && y >= 2 && y < c_height - 2) ) {
						// Yes! Are there any of these wall blocks within 2 cells?
						var clear:Bool = true;
						for ( tx in x - 2 ... x + 3 ) {
							for ( ty in y - 2 ... y + 3 ) {
								if ( m_screen[ty * c_width + tx] != 0 ) {
									// Unfortunately, yes... we can't place here.
									clear = false;
								}
							}
						}
						
						if ( clear ) {
							// We can build here. Set it and move on to doing the next.
							m_screen[idx] = -5;
							m_freeSpaces--;
							break;
						}
					}
				}
			}
		}
		
		// Set up snake-related data.
		m_headX = 6;
		m_headY = Math.floor(0.5 * c_height);

		m_snakeLength = 5; // Start 5 units long - your minimum length.
		m_blood = 0; // Start not bleeding.
		
		// Add the snake's body to the level.
		m_screen[m_headY * c_width + m_headX] = 1;
		m_screen[m_headY * c_width + m_headX - 1] = 2;
		m_screen[m_headY * c_width + m_headX - 2] = 3;
		m_screen[m_headY * c_width + m_headX - 3] = 4;
		m_screen[m_headY * c_width + m_headX - 4] = 5;
		m_freeSpaces -= 5;

		// Add apples.
		addApple();
		addApple();
		addApple();
		addApple();
		addApple();
		
		m_dir = 0; // 0 = right, 1 = down, 2 = left, 3 = up
		m_prevDir = 0;
		m_nonAdvanceFrames = 0;
		
		m_score = 0;
		
		m_gameStart = false;
		m_gameOver = false;
		m_gameOverTime = 3.5; // Seconds before the game restarts on game over.
		
		// Initial state of input.
		m_setDir = false;
		m_queuedDir = -1;
	}
	
	public function activate()
	{
	}
	
	public function deactivate()
	{
	}

	private function addApple() {
		var freeSpacesLeft:Int = m_freeSpaces;
		
		// We know how many free spaces there are in the screen, so iterate over all spaces
		// rolling the dice for each one to check the odds of using it vs. the free spaces we
		// haven't got to yet.
		for ( i in 0 ... c_height * c_width ) {
			if ( m_screen[i] == 0 || m_screen[i] == -4 ) {
				if ( Math.random() < 1 / freeSpacesLeft ) {
					m_screen[i] = -1;
					m_freeSpaces--;
					return;
				}
				
				freeSpacesLeft--;
			}
		}
		
		// If we get here, m_freeSpaces was larger than the actual number of free spaces.
		// Throw an exception.
		throw( "Error in m_freeSpaces: " + freeSpacesLeft );
	}

	public function advance( _dt:Float, _context:AdvanceContext ) : Void
	{
		// Handle input.
		
		// Toggle blur.
		if ( _context.m_input.wasKeyPressed( GameInput.c_keyQ ) ) {
			s_blur = !s_blur;
		}
		
		// Update movement direction.
		var updateDir:Int = -1;
		if ( _context.m_input.wasKeyPressed( GameInput.c_keyArrowRight ) || _context.m_input.wasKeyPressed( GameInput.c_keyD ) ) {
			updateDir = 0;
		} else if ( _context.m_input.wasKeyPressed( GameInput.c_keyArrowDown ) || _context.m_input.wasKeyPressed( GameInput.c_keyS ) ) {
			updateDir = 1;
		} else if ( _context.m_input.wasKeyPressed( GameInput.c_keyArrowLeft ) || _context.m_input.wasKeyPressed( GameInput.c_keyA ) ) {
			updateDir = 2;
		} else if ( _context.m_input.wasKeyPressed( GameInput.c_keyArrowUp ) || _context.m_input.wasKeyPressed( GameInput.c_keyW ) ) {
			updateDir = 3;
		}
		
		// If we haven't started the game yet, don't advance. Check for input.
		if ( !m_gameStart ) {
			if ( updateDir != -1 ) {
				m_gameStart = true;
			} else {
				return;
			}
		}
		
		// If the game's over, don't advance. If the timer's expired, set a high score and restart.
		if ( m_gameOver ) {
			m_gameOverTime -= _dt;
			if ( m_gameOverTime < 0.0 ) {
				if ( m_score > s_highScore ) {
					s_highScore = m_score;
				}
				
				_context.m_sceneManager.changeScene( new SceneSnake() );
			}
			
			return;
		}		
		
		// Don't let players double back.
		if ( Math.abs( m_dir - updateDir ) == 2 ) {
			updateDir = -1;
		}
		
		// Have we set a new direction for the next advance frame? Should we queue a direction order?
		if ( updateDir != -1 ) {
			if ( !m_setDir ) {
				m_dir = updateDir;
				m_setDir = true;
			} else {
				m_queuedDir = updateDir;
			}
		}
		
		// Delay loop - only do a proper advance every 6 frames.
		if ( m_nonAdvanceFrames > 0 ) {
			m_nonAdvanceFrames--;
			return;
		} else {
			m_nonAdvanceFrames = 5;
		}

		// From here on, we're doing the actual game logic! ----------------------------------
		
		// Store the direction we're going to move this frame for future reference.
		m_prevDir = m_dir;
		
		// Run through the array and degrade all old snake parts.
		for ( i in 0 ... c_width * c_height ) {
			if ( m_screen[i] > 0 ) {
				m_screen[i]++; // Every live snake part gets a tick older.

				if ( m_screen[i] > m_snakeLength ) {
					// Live snake parts that are too old disappear.
					m_screen[i] = 0;
					if ( m_blood > 0 ) {
						m_screen[i] = -4;
					}
					m_freeSpaces++;
				}
			} else if ( m_screen[i] == 0 || m_screen[i] == -4 || m_screen[i] == -1 ) {
				// Empty, blood trail or apple.
				
				// Is this cell reachable? If it's a single-space-wide dead end, we'll never be able to
				// get at it without dying! Eliminate it and replace it with a wall.
				
				// Count neighbours.
				var ni:Int;
				var count:Int = 0;

				ni = i - 1;
				if ( ni >= 0 && (i % c_width != 0) && (m_screen[ni] >= 0 || m_screen[ni] == -4 || m_screen[ni] == -1) ) {
					count++;
				}
				ni = i - c_width;
				if ( ni >= 0 && (m_screen[ni] >= 0 || m_screen[ni] == -4 || m_screen[ni] == -1) ) {
					count++;
				}
				ni = i + 1;
				if ( ni < c_width*c_height && (ni % c_width != 0) && (m_screen[ni] >= 0 || m_screen[ni] == -4 || m_screen[ni] == -1) ) {
					count++;
				}
				ni = i + c_width;
				if ( ni < c_width*c_height && (m_screen[ni] >= 0 || m_screen[ni] == -4 || m_screen[ni] == -1) ) {
					count++;
				}
				
				if ( count <= 1 ) {
					// This cell only has one empty neighbour. Fill it in.
					var wasApple:Bool = (m_screen[i] == -1);
					
					m_screen[i] = -5; // wall
					
					if ( wasApple ) {
						addApple(); // Maintain five apples at all times.
					} else {
						m_freeSpaces--;
					}
				}
			}
		}
		
		// Move the head position to a new spot, according to our movement direction.
		switch ( m_dir ) {
			case 0: m_headX++;
			case 1: m_headY++;
			case 2: m_headX--;
			case 3: m_headY--;
		}
		
		// Bounds check that move.
		if ( m_headX < 0 || m_headX >= c_width || m_headY < 0 || m_headY >= c_height ) {
			// Out of bounds.
			for ( i in 0 ... c_width * c_height ) {
				if ( m_screen[i] == 2 ) {
					m_screen[i] = -3; // Set the snake section behind the head to a fail block.
				}
			}				

			SoundPlayer.get().playSound( new WavWall(), 0.7 );
			
			m_gameOver = true;
			return;
		}

		// Have we hit something?
		if ( m_screen[m_headY * c_width + m_headX] == -1 ) {
			// Hit an apple.
			SoundPlayer.get().playSound( new WavApple(), 0.4 + Math.random()*0.3 );

			addApple();

			m_freeSpaces++; // This apple's being eliminated, so free up a space.
			
			++m_snakeLength; // Grow the snake.
//			m_score++; // Restore this for conventional Snake gameplay
		} else if ( m_screen[m_headY * c_width + m_headX] > 0 ) {
			// Hit a snake part.
			SoundPlayer.get().playSound( new WavCannibal(), 0.7 );

			// Start bleeding from our fresh-cut tail.
			m_blood = Math.floor( 15 + 30 * Math.random() );
			
			m_freeSpaces++;

			// Kill off every snake section older than this section.
			for ( i in 0 ... c_width * c_height ) {
				if ( m_screen[i] > m_screen[m_headY * c_width + m_headX] ) {
					m_screen[i] = -2;
					m_score++;
					m_snakeLength--;
					
					m_blood += 2; // Increase our blood timer for every section cut...
				}
			}
			
//			m_gameOver = true; // Restore this for conventional Snake gameplay
//			return; // Restore this for conventional Snake gameplay
		} else if ( m_screen[m_headY * c_width + m_headX] == -2 ) {
			// Hit a dead snake section. Game over.
			m_screen[m_headY * c_width + m_headX] = -3; // Show red fail block.
			SoundPlayer.get().playSound( new WavWall(), 0.7 );
			m_gameOver = true;
			return;
		} else if ( m_screen[m_headY * c_width + m_headX] == -5 ) {
			// Hit an internal wall. Game over.
			m_screen[m_headY * c_width + m_headX] = -3; // Show red fail block.
			SoundPlayer.get().playSound( new WavWall(), 0.7 );
			m_gameOver = true;
			return;
		}
		
		// Finally, create the new snake section.
		m_screen[m_headY * c_width + m_headX] = 1;
		m_freeSpaces--;

		// Reduce our blood timer.
		if ( m_blood > 0 )
		{
			m_blood--;
		}
		
		// If we have a queued direction order, apply it now.
		if ( m_queuedDir != -1 ) {
			m_dir = m_queuedDir;
			m_setDir = true;
		} else {
			m_setDir = false;
		}
		
		m_queuedDir = -1;
	}
	
	public function render( _targetBD:BitmapData ) : Void
	{
		// Clear the screen.
		_targetBD.fillRect( _targetBD.rect, 0x000000 );

		// Draw the boundaries.
		_targetBD.fillRect( new Rectangle( 0, 0, 10, 150 ), 0x707070 );
		_targetBD.fillRect( new Rectangle( 190, 0, 10, 150 ), 0x707070 );
		_targetBD.fillRect( new Rectangle( 10, 0, 180, 10 ), 0x707070 );
		_targetBD.fillRect( new Rectangle( 10, 140, 180, 10 ), 0xA0A0A0 );
		
		// Render everything, except for the snake's head and tail - we want to render
		// those last so they overlap other snake parts. Store the location of the tail
		// when we find it (we know where the head is)
		var tailHighest:Int = 0;
		var tailX:Int = -1;
		var tailY:Int = -1;
		
		var rect:Rectangle = new Rectangle( 0, 0, c_tileSize, c_tileSize );
		for ( x in 0 ... c_width ) {
			for ( y in 0 ... c_height ) {
				var screenIdx = y * c_width + x;

				rect.x = 10 + c_tileSize * x;
				rect.y = 10 + c_tileSize * y;
				
				if ( m_screen[screenIdx] == -1 ) {
					_targetBD.fillRect( rect, 0x80ff80 ); // Apple
				} else if ( m_screen[screenIdx] == -2 ) {
					_targetBD.fillRect( rect, 0x806010 ); // Dead snake part
				} else if ( m_screen[screenIdx] == -3 ) {
					_targetBD.fillRect( rect, 0xa00000 ); // Red fail block
				} else if ( m_screen[screenIdx] == -4 ) {
					_targetBD.fillRect( rect, 0x200000 ); // Blood trail
				} else if ( m_screen[screenIdx] == -5 ) {
					_targetBD.fillRect( rect, 0x707070 ); // Wall
				} else if ( m_screen[screenIdx] == 1 ) {
					// The snake's head. Don't render it yet.
				} else if ( m_screen[screenIdx] > 1 ) {
					// Another snake part. Render in brown.
					if ( m_screen[screenIdx] > tailHighest ) { // If this looks like the tail, remember it.
						tailHighest = m_screen[screenIdx];
						tailX = x;
						tailY = y;
					}
					_targetBD.fillRect( rect, 0xffa040 );
				}
			}
		}

		// Draw the tail.
		rect.x = 10 + c_tileSize * tailX;
		rect.y = 10 + c_tileSize * tailY;

		// Draw the background *under* the tail as it's leaving the square.
		if ( m_blood > 0 ) {
			_targetBD.fillRect( rect, 0x200000 );
		} else {
			_targetBD.fillRect( rect, 0x000000 );
		}
		
		// Hunt for the next-oldest snake section next to the tail. The tail
		// will be moving in that direction.
		if ( tailHighest == m_snakeLength ) {
			var bestTailDir = 0;
			var bestTailDirAge = 0;

			if ( tailX < c_width - 1 && m_screen[tailY * c_width + tailX + 1] > bestTailDirAge ) {
				bestTailDir = 0;
				bestTailDirAge = m_screen[tailY * c_width + tailX + 1];
			}
			if ( tailY < c_height - 1 && m_screen[(tailY + 1) * c_width + tailX] > bestTailDirAge ) {
				bestTailDir = 1;
				bestTailDirAge = m_screen[(tailY + 1) * c_width + tailX];
			}
			if ( tailX > 0 && m_screen[tailY * c_width + tailX - 1] > bestTailDirAge ) {
				bestTailDir = 2;
				bestTailDirAge = m_screen[tailY * c_width + tailX - 1];
			}
			if ( tailY > 0 && m_screen[(tailY - 1) * c_width + tailX] > bestTailDirAge ) {
				bestTailDir = 3;
				bestTailDirAge = m_screen[(tailY - 1) * c_width + tailX];
			}

			// Smoothly move the tail from its own location over to the next-oldest snake section's location.
			switch ( bestTailDir ) {
				case 0: rect.x += c_tileSize * (1 - m_nonAdvanceFrames / 6);
				case 1: rect.y += c_tileSize * (1 - m_nonAdvanceFrames / 6);
				case 2: rect.x -= c_tileSize * (1 - m_nonAdvanceFrames / 6);
				case 3: rect.y -= c_tileSize * (1 - m_nonAdvanceFrames / 6);
			}
		}
		
		if ( m_blood > 0 ) {
			_targetBD.fillRect( rect, 0x802010 );
		} else {
			_targetBD.fillRect( rect, 0xb08020 );
		}
		
		// Draw the head. Smoothly move into place from the previous square.
		if ( !m_gameOver ) {
			rect.x = 10 + c_tileSize * m_headX;
			rect.y = 10 + c_tileSize * m_headY;
			switch ( m_prevDir ) {
				case 0: rect.x -= c_tileSize * m_nonAdvanceFrames / 6;
				case 1: rect.y -= c_tileSize * m_nonAdvanceFrames / 6;
				case 2: rect.x += c_tileSize * m_nonAdvanceFrames / 6;
				case 3: rect.y += c_tileSize * m_nonAdvanceFrames / 6;
			}
			_targetBD.fillRect( rect, 0xffd090 );
		}

		// Render UI text.
		FontData.get().getFont().renderText( _targetBD, "CORPSE SNAKE", 100, 1, 0.5 );
		
		// Show our score in gold if it beats the current high score; otherwise, white.
		if ( m_score > s_highScore ) {
			FontData.get().getGoldFont().renderText( _targetBD, "SCORE: " + m_score, 11, 141, 1.0 );
		} else {
			FontData.get().getFont().renderText( _targetBD, "SCORE: " + m_score, 11, 141, 1.0 );
		}
		FontData.get().getFont().renderText( _targetBD, "HIGH SCORE: " + s_highScore, 188, 141, 0.0 );
		
		// Startup text.
		if ( !m_gameStart ) {
			FontData.get().getFont().renderText( _targetBD, "A GAME BY RANDOMNINE", 100, 31, 0.5 );
			FontData.get().getGoldFont().renderText( _targetBD, "BITE OFF TAIL FOR POINTS", 100, 41, 0.5 );
			FontData.get().getFont().renderText( _targetBD, "UP / DOWN / LEFT / RIGHT", 100, 101, 0.5 );
			FontData.get().getFont().renderText( _targetBD, "or W / S / A / D", 100, 111, 0.5 );
		}
		
		// Game over text.
		if ( m_gameOver ) {
			FontData.get().getFont().renderText( _targetBD, "GAME OVER", 100, 101, 0.5 );
		}

		// Finally, apply blur filter. (We could apply this before adding the text, but I prefer it this way.)
		if ( s_blur ) {
			m_filter.shader.data.strength.value = [1.0];
			_targetBD.applyFilter( _targetBD, _targetBD.rect, new Point(0, 0), m_filter );
		}
	}	
}
